/* wptKeyEdit.h
 *	Copyright (C) 2003, 2004 Timo Schulz
 *
 * This file is part of WinPT.
 *
 * WinPT is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation; either version 2 
 * of the License, or (at your option) any later version.
 *  
 * WinPT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with WinPT; if not, write to the Free Software Foundation, 
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 */

#ifndef WPT_KEY_EDIT_H
#define WPT_KEY_EDIT_H

/* IDs for all possible gpg edit key commands. */
enum gpg_editkey_t {
    GPG_EDITKEY_SIGN    =  1,
    GPG_EDITKEY_LSIGN   =  2,
    GPG_EDITKEY_NRSIGN  = 19,
    GPG_EDITKEY_NRLSIGN = 22,
    GPG_EDITKEY_TSIGN   = 20,
    GPG_EDITKEY_TRUST   =  3,
    GPG_EDITKEY_ADDUID  =  4,
    GPG_EDITKEY_DELUID  =  5,
    GPG_EDITKEY_DELKEY  =  6,
    GPG_EDITKEY_ADDKEY  =  7,
    GPG_EDITKEY_PASSWD  =  8,
    GPG_EDITKEY_PRIMARY =  9,
    GPG_EDITKEY_EXPIRE  = 10,
    GPG_EDITKEY_REVSIG  = 11,
    GPG_EDITKEY_REVKEY  = 12,
    GPG_EDITKEY_REVOKE  = 13, /* revoke the whole key */
    GPG_EDITKEY_ADDREV  = 14,
    GPG_EDITKEY_ADDPHOTO= 15,
    GPG_EDITKEY_ENABLE  = 16,
    GPG_EDITKEY_DISABLE = 17,
    GPG_EDITKEY_SETPREF = 18,
    GPG_EDITKEY_DELSIG  = 21,
    GPG_EDITKEY_KEYSERV = 23,
    GPG_EDITKEY_REVUID  = 24,
    GPG_EDITKEY_CLEAN	= 25
};

/* Reasons for revocations. */
enum gpg_revoke_t {
    GPG_REVOKE_INSECURE   = 1,
    GPG_REVOKE_SUPERSEDED = 2,
    GPG_REVOKE_NOUSED     = 3
};


/* Possible results for the operation. */
enum gpg_edit_result_t {
    GPG_EDITRES_ALREADY_SIGNED = 1,
    GPG_EDITRES_BAD_PASSPHRASE = 2
};

/* Contain information about designated revokers. */
struct gpg_desig_rev_s {
    struct gpg_desig_rev_s *next;
    gpgme_pubkey_algo_t pubkey_algo;
    char fpr[41];
};
typedef struct gpg_desig_rev_s *gpg_desig_rev_t;

/* Context to hold all user-ID information. */
struct gpg_user_id_info_s {
    struct gpg_user_id_info_s *next;
    struct {
	unsigned int revoked:1;
	unsigned int invalid:1;
	unsigned int primary:1;
	unsigned int mdc:1;
	unsigned int no_ks_modify:1;
    } flags;
    int  validity;
    char *name;
    char *email;
    char *prefs;    
    int  index;
};
typedef struct gpg_user_id_info_s *gpg_uid_info_t;

/* Class to abstract the gpg edit key interface. */
class GpgKeyEdit {
private:
    gpgme_key_t key;
    gpgme_ctx_t ctx;
    int resval;
    int type;

public:
    
    int cnt;
    int cmd_sent;
    const char *pass;
    const char *new_pass;
    int sig_index;
    int key_index;
    int uid_index;
    const char *url;
    int trust_id;
    char *name, *cmt, *email;
    const char *exp_date; /* XXX */
    gpgme_pubkey_algo_t pubkey_algo;
    int pubkey_size;
    int valid; /* XXX */
    int flags;
    int sig_class;
    int reason;

public:
    GpgKeyEdit (gpgme_key_t key);
    GpgKeyEdit (const char *keyid);
    ~GpgKeyEdit (void);

    bool isValid (void);
    gpgme_key_t getKey (void);
    void reset (void);

    void setPassphrase (const char *pass);
    void setKey (gpgme_key_t key);
    void setKeyID (const char *keyid);
    void setLocalUser (gpgme_key_t locusr);
    void setCallback (gpgme_progress_cb_t cb, void *cb_value);
    int  getType (void);
    void setResult (int val);
    int  getResult (void);

    gpgme_error_t signKey (int mode, int sig_class, const char *exp_date);
    gpgme_error_t setTrust (gpgme_validity_t trust);
    gpgme_error_t addUserid (const char *name, const char *cmt, 
			     const char *email);
    gpgme_error_t delUserid (int index);
    gpgme_error_t delKey (int index);
    gpgme_error_t addSubkey (gpgme_pubkey_algo_t pubkey_algo, 
			     unsigned int nbits, long valid);
    gpgme_error_t changePassphrase (const char *new_pass, int allow_empty);
    gpgme_error_t setPrimaryUserid (int uid_indx);
    gpgme_error_t setKeyExpireDate (int key_index, long exp_timestamp, 
				    bool exp_days);
    gpgme_error_t revokeUserid (int uid_indx);
    gpgme_error_t revokeSignature (int uid_index, int sig_index);
    gpgme_error_t revokeSubkey (int key_index, int reason, const char *cmt);
    gpgme_error_t addDesignatedRevoker (const char *uid);
    gpgme_error_t addPhotoid (const char *jpg_file);
    gpgme_error_t enable (void);
    gpgme_error_t disable (void);
    gpgme_error_t setUseridPreferences (int uid_index, const char *new_prefs);
    gpgme_error_t delUseridSignature (int uid_index, int sig_index);
    gpgme_error_t setPreferredKeyserver (int uid_index, const char *url);
    gpgme_error_t cleanKey (void);

    gpgme_error_t getUseridInfo (gpg_uid_info_t *r_uinf);

    gpgme_error_t getDesignatedRevoker (gpg_desig_rev_t *r_rev);
};

BOOL keyedit_add_subkey (winpt_key_t k, HWND dlg, listview_ctrl_t lv);
BOOL keyedit_add_userid (winpt_key_t k, HWND dlg, listview_ctrl_t lv);
BOOL keyedit_add_revoker (winpt_key_t k, HWND dlg);
BOOL keyedit_add_photo (winpt_key_t k, HWND dlg );
BOOL keyedit_change_passwd (winpt_key_t k, HWND dlg);
BOOL keyedit_set_pref_keyserver (winpt_key_t k, HWND dlg);

gpgme_error_t gpg_editkey (gpgme_ctx_t ctx, gpgme_key_t key, GpgKeyEdit *ek);
void	      gpg_uid_info_release (gpg_uid_info_t inf);
void	      gpg_desig_rev_release (gpg_desig_rev_t rev);

#endif /*WPT_KEY_EDIT_H*/
