/* wptCommonDlg.cpp
 *	Copyright (C) 2004, 2005 Timo Schulz
 *
 * This file is part of WinPT.
 *
 * WinPT is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * WinPT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with WinPT; if not, write to the Free Software Foundation, 
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <windows.h>

#include "resource.h"

#include "wptTypes.h"
#include "wptW32API.h"
#include "wptVersion.h"
#include "wptGPG.h"
#include "wptCommonCtl.h"
#include "wptContext.h"


/* check if the given buf contains a valid URL. */
static int
check_URL (const char *buf)
{
    /* XXX: be more strict */
    if (!strchr (buf, ':'))
	return -1;
    if (!strstr (buf, "//"))
	return -1;
    return 0;
}


static BOOL CALLBACK
http_dlg_proc (HWND dlg, UINT msg, WPARAM wparam, LPARAM lparam)
{
    static struct URL_ctx_s * ctx;
    int n;

    switch (msg) {
    case WM_INITDIALOG:
	ctx = (struct URL_ctx_s*)lparam;
	if (!ctx)
	    BUG (0);
	if (ctx->desc != NULL)
	    SetWindowText (dlg, ctx->desc);
	if (ctx->title != NULL)
	    SetDlgItemText (dlg, IDC_HTTP_TITLE, ctx->title);
	else
	    SetDlgItemText (dlg, IDC_HTTP_TITLE, _("Enter URL to retrieve the public key"));
	SetDlgItemText (dlg, IDCANCEL, _("&Cancel"));
	SetWindowText (dlg, _("HTTP Key Import"));
	SetForegroundWindow (dlg);
	break;

    case WM_COMMAND:
	switch (LOWORD (wparam)) {
	case IDOK:
	    n = GetDlgItemText (dlg, IDC_HTTP_URL, ctx->url, DIM (ctx->url)-1);
	    if (ctx->check && (!n || check_URL (ctx->url))) {
		msg_box (dlg, _("Please enter a valid URL."), "HTTP", MB_ERR);
		return FALSE;
	    }
	    ctx->cancel = 0;
	    EndDialog (dlg, TRUE);
	    break;

	case IDCANCEL:
	    ctx->cancel = 1;
	    EndDialog (dlg, FALSE);
	    break;
	}
    }

    return FALSE;
}


/* Common dialog to request a URL from the user. 
   Return value is the URL context. */
void*
get_http_file_dlg (HWND root)
{
    struct URL_ctx_s *ctx;

    ctx = new URL_ctx_s;
    if (!ctx)
	BUG (0);
    memset (ctx, 0, sizeof *ctx);
    ctx->check = 1;	
    DialogBoxParam (glob_hinst, (LPCTSTR)IDD_WINPT_HTTP, root,
		    http_dlg_proc, (LPARAM)ctx);
    return ctx;
}

/* Wrapper to request a preferred keyserver from the user.
   Return value is the URL context. */
void*
get_keyserver_URL_dlg (HWND root)
{
    struct URL_ctx_s *ctx;

    ctx = new URL_ctx_s;
    if (!ctx)
	BUG (0);
    memset (ctx, 0, sizeof *ctx);
    ctx->check = 1;
    ctx->desc = _("Key Edit");
    ctx->title = _("Enter preferred keyserver URL"); 
    DialogBoxParam (glob_hinst, (LPCTSTR)IDD_WINPT_HTTP, root,
		    http_dlg_proc, (LPARAM)ctx);
    return ctx;
}


/* Generic text input dialog. */
char*
get_input_dialog (HWND root, const char *title, const char *desc)
{
    struct URL_ctx_s ctx;

    memset (&ctx, 0, sizeof (ctx));
    ctx.title = title;
    ctx.desc = desc;
    DialogBoxParam (glob_hinst, (LPCTSTR)IDD_WINPT_HTTP, root,
		    http_dlg_proc, (LPARAM)&ctx);
    if (ctx.cancel || strlen (ctx.url) == 0)
	return NULL;
    return m_strdup (ctx.url);
}
