/* wptKeyManagerDlg.cpp - WinPT Key Manager
 *	Copyright (C) 2000-2005 Timo Schulz
 *	Copyright (C) 2004 Andreas Jobs
 *
 * This file is part of WinPT.
 *
 * WinPT is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * WinPT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with WinPT; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <windows.h>
#include <commctrl.h>

#include "resource.h"
#include "wptTypes.h"
#include "wptGPG.h"
#include "wptCommonCtl.h"
#include "wptContext.h" /* for passphrase_s */
#include "wptDlgs.h"
#include "wptW32API.h"
#include "wptVersion.h"
#include "wptKeylist.h"
#include "wptNLS.h"
#include "wptErrors.h"
#include "wptHotkey.h"
#include "wptKeyManager.h"
#include "wptKeyserver.h"
#include "wptKeyEdit.h"
#include "wptRegistry.h"

/* Name and ID of the separator window. */
#define KM_SEPARATOR_ID		       10000
#define WINDOWCLASS_SEPARATOR_CHILD "WINPT_SEP_CHILD"

/* Virtual key codes. */
#ifndef VK_F
#define VK_F 70
#endif
#ifndef VK_A
#define VK_A 65
#endif
#ifndef VK_C
#define VK_C 67
#endif
#ifndef VK_P
#define VK_P 80
#endif

static subclass_s keylist_proc;

HIMAGELIST glob_imagelist;

struct km_info {
    /* Window positions */
    int pos_x, pos_y;
    int ypos_sep;
    int ypercent_sep;

    HWND hwnd_sep;
    HWND toolbar;
    HWND statbar;

    listview_ctrl_t lv;
    int		    lv_idx;
    int keylist_sortby;
};

/* Toolbar button structure. */
struct mybuttons {
    long icon;
    long command;
    char *text;
};

struct mybuttons myb[] = {
    {IMI_KEY_DELETE, ID_KEYMISC_DELETE, "Delete"},
    {IMI_KEY_PROPS,  ID_KEYMISC_PROPS,	"Properties"},
    {IMI_KEY_SIGN,   ID_KEYMISC_SIGN,	"Sign"},
    {IMI_KEY_IMPORT, ID_KEYMISC_IMPORT, "Import"},
    {IMI_KEY_EXPORT, ID_KEYMISC_EXPORT, "Export"},
};


/* Subclass the keylist listview control to allow extended commands. */
static BOOL CALLBACK
keylist_subclass_proc (HWND dlg, UINT msg, WPARAM wparam, LPARAM lparam)
{
    struct listview_ctrl_s lv;

    switch (msg) {
    case WM_LBUTTONDBLCLK:
        send_cmd_id (keylist_proc.dlg, ID_KEYMISC_PROPS);
        break;
        
    case WM_KEYUP:
        int virt_key = (int)wparam;
	switch (virt_key) {
	case VK_SPACE:
            send_cmd_id (keylist_proc.dlg, ID_KEYMISC_PROPS);
            break;
        
	case VK_DELETE:
            send_cmd_id (keylist_proc.dlg, ID_KEYMISC_DELETE);
            break;
        
	case VK_INSERT:
	    send_cmd_id (keylist_proc.dlg, ID_KEYMISC_PASTE);
	    break;

	case VK_A:
	    if (GetAsyncKeyState (VK_CONTROL)) {
		lv.ctrl = GetDlgItem (keylist_proc.dlg, IDC_KEYMISC_KEYLIST);
		listview_select_all (&lv);
	    }
	    break;
	    
	case VK_C:
	    if (GetAsyncKeyState (VK_CONTROL)) {
		lv.ctrl = GetDlgItem (keylist_proc.dlg, IDC_KEYMISC_KEYLIST);
		km_clip_export (keylist_proc.dlg, &lv);
	    }
	    break;

	case VK_P:
	    if (GetAsyncKeyState (VK_CONTROL))		
		km_clip_import (keylist_proc.dlg);
	    break;

	case VK_F:
	    if (GetAsyncKeyState (VK_CONTROL)) {
		lv.ctrl = GetDlgItem (keylist_proc.dlg, IDC_KEYMISC_KEYLIST);
		km_find_key (dlg, &lv);
	    }
	    break;
	}
        break;
    }
    
    return CallWindowProc (keylist_proc.old, dlg, msg, wparam, lparam);
}


static HWND
load_toolbar (HWND dlg, struct km_info *kmi)
{
    HWND tbwnd;
    TBSAVEPARAMS tbsp;
    TBBUTTON tb_buttons[] = {
        /*{imagelist_getindex(IMI_EXIT),       ID_KEYMISC_QUIT,   TBSTATE_ENABLED, TBSTYLE_BUTTON, 0, 0L, 0},*/
	{imagelist_getindex(IMI_KEY_DELETE), ID_KEYMISC_DELETE, TBSTATE_ENABLED, TBSTYLE_BUTTON, {0}, 0L, 0},
        {imagelist_getindex(IMI_KEY_PROPS),  ID_KEYMISC_PROPS,  TBSTATE_ENABLED, TBSTYLE_BUTTON, {0}, 0L, 0},
	{imagelist_getindex(IMI_KEY_SIGN),   ID_KEYMISC_SIGN,   TBSTATE_ENABLED, TBSTYLE_BUTTON, {0}, 0L, 0},
        {0,                                  0,                 0,               TBSTYLE_SEP,    {0}, 0L, 0},
	{imagelist_getindex(IMI_KEY_IMPORT), ID_KEYMISC_IMPORT, TBSTATE_ENABLED, TBSTYLE_BUTTON, {0}, 0L, 0},
        {imagelist_getindex(IMI_KEY_EXPORT), ID_KEYMISC_EXPORT, TBSTATE_ENABLED, TBSTYLE_BUTTON, {0}, 0L, 0},
     };
    
    tbwnd = CreateWindowEx (0, TOOLBARCLASSNAME, NULL,
			    WS_CHILD|TBSTYLE_TOOLTIPS|TBSTYLE_FLAT|CCS_ADJUSTABLE,
			    0, 0, 0, 0, dlg, (HMENU)IDR_WINPT_KMTB, glob_hinst, NULL);
    if (tbwnd) {
	SendMessage (tbwnd, TB_BUTTONSTRUCTSIZE, (WPARAM) sizeof(TBBUTTON), 0);
        SendMessage (tbwnd, TB_SETIMAGELIST, 0, (LPARAM)glob_imagelist);
	        SendMessage (tbwnd, TB_AUTOSIZE, 0, 0);
        ShowWindow (tbwnd, SW_SHOW);

        /* Restore toolbar */
        tbsp.hkr = HKEY_CURRENT_USER;
        tbsp.pszSubKey = "Software\\WinPT";
        tbsp.pszValueName = "KM_toolbar";
        if (SendMessage(tbwnd, TB_SAVERESTORE, FALSE, (LPARAM)&tbsp ) == 0)
            SendMessage (tbwnd, TB_ADDBUTTONS, sizeof(tb_buttons) / sizeof(tb_buttons[0]), 
			 (LONG)&tb_buttons[0]);
     }
     return tbwnd;
}


/* Center window @dlg. */
static void
do_center_window (HWND dlg, struct km_info *kmi)
{
    RECT rect;
    char *p;
    int pos_x = 0, pos_y = 0;
        
    /* Find bottom of keylist */
    GetWindowRect (GetDlgItem(dlg, IDC_KEYMISC_KEYLIST), &rect);
    MapWindowPoints (NULL, dlg, (POINT*)&rect, 2);

    kmi->ypos_sep = rect.bottom;

    p = get_reg_entry (HKEY_CURRENT_USER, "Software\\WinPT", "KM_Pos_X");
    if (p && !strcmp (p, " ")) {
	free_if_alloc (p);	
	center_window (dlg, NULL);
	return;
    }
    else if (p)
	pos_x = atol (p);

    p = get_reg_entry (HKEY_CURRENT_USER, "Software\\WinPT", "KM_Pos_Y");
    if (p && !strcmp (p, " ")) {
	free_if_alloc (p);
	center_window (dlg, NULL);
	return;
    }
    else if (p)
	pos_y = atol (p);

    if (!pos_y && !pos_x) {
	center_window (dlg, NULL);
	return;
    }
    
    if (pos_x < 0 || pos_y < 0)
	pos_x = pos_y = 0;
    if (pos_x > GetSystemMetrics (SM_CXSCREEN) 
	|| pos_y > GetSystemMetrics (SM_CYSCREEN)) {
	pos_x = pos_y = 0;
    }
    GetClientRect (dlg, &rect);
    MoveWindow (dlg, pos_x, pos_y, rect.right, rect.bottom, TRUE);
}


/* Resize the key manager window with the information from @kmi. */
static void
do_resize_window (HWND dlg, struct km_info *kmi)
{
    HWND h;
    RECT rclient, rect;
    BOOL bRepaint = FALSE;

    /* Get rect of client area and make life easier */
    GetClientRect (dlg, &rclient);

    /* Move toolbar to the top of the window */
    if (kmi->toolbar) {
        GetWindowRect (kmi->toolbar, &rect);
        ScreenToClient (dlg, (POINT*)&rect);
        ScreenToClient (dlg, (POINT*)&(rect.right));

        rclient.top += rect.bottom - rect.top;
        MoveWindow (kmi->toolbar, 0, 0, rclient.right - rclient.left, 
		    rect.bottom - rect.top, bRepaint);
    }

    /* Move statusbar to the bottom of the window */
    if (kmi->statbar) {
        GetWindowRect (kmi->statbar, &rect);
        ScreenToClient (dlg, (POINT*)&rect);
        ScreenToClient (dlg, (POINT*)&(rect.right));

        rclient.bottom -= rect.bottom - rect.top;
        MoveWindow (kmi->statbar, 0, rclient.bottom,
		    rclient.right - rclient.left,
		    rect.bottom - rect.top, bRepaint);
    }

    /* Obtain separator information and move it to the desired posistion */
    if (kmi->ypercent_sep)
        kmi->ypos_sep = (rclient.bottom - rclient.top) * kmi->ypercent_sep / 100;
    else
        kmi->ypercent_sep = kmi->ypos_sep * 100 / (rclient.bottom - rclient.top);
        
    /* Don't move away */
    if (kmi->ypos_sep+5 > rclient.bottom)
        kmi->ypos_sep = rclient.bottom - 5;
    if (kmi->ypos_sep < rclient.top)
        kmi->ypos_sep = rclient.top;
    MoveWindow (kmi->hwnd_sep, 0, kmi->ypos_sep, 
		(rclient.right - rclient.left), 5, bRepaint);
        
    /* Place the keylist above the separator */
    h = GetDlgItem (dlg, IDC_KEYMISC_KEYLIST);
    MoveWindow (h, rclient.left, rclient.top, rclient.right - rclient.left, 
		kmi->ypos_sep - rclient.top, bRepaint);
    rclient.top = kmi->ypos_sep + 5 + 8;

    /* Place the group text and the group box below the separator */
    h = GetDlgItem (dlg, IDC_KEYMISC_GTEXT);
    MoveWindow (h, rclient.left, rclient.top, 100, 14, bRepaint);
    rclient.top += 18;

    h = GetDlgItem (dlg, IDC_KEYMISC_GROUP);
    MoveWindow (h, rclient.left, rclient.top, rclient.right - rclient.left, 
		(rclient.bottom < rclient.top) ? 
		0 : rclient.bottom - rclient.top, bRepaint);
    
    /* Repaint the whole thing */
    InvalidateRect (dlg, NULL, TRUE);
}


/* Return true if the clipboard contains an OpenPGP key. */
static bool
clip_contains_pgpkey (void)
{
    char *ctxt;
    bool val = false;

    ctxt = get_clip_text (NULL);
    if (!ctxt || strlen (ctxt) < 512)
	val = false;
    else if (strstr (ctxt, "BEGIN PGP") && strstr (ctxt, "KEY BLOCK") &&
	     strstr (ctxt, "END PGP"))
	val = true;
    free_if_alloc (ctxt);
    return val;
}


/* Show a mini popup menu to import keys. */
static void
do_create_minpopup (HWND dlg)
{
    HMENU hm;
    MENUITEMINFO mi;
    char * s;
    POINT p;
    
    if (gpg_read_only || !clip_contains_pgpkey ())
	return;
    hm = CreatePopupMenu ();
    if (!hm)
	BUG( NULL );
    memset (&mi, 0, sizeof mi);
    mi.cbSize = sizeof mi;
    s = (char *)_("Paste Key from Clipboard");
    mi.fType = MF_STRING;
    mi.dwTypeData = s;
    mi.cch = strlen (s);
    mi.fMask = MIIM_DATA | MIIM_ID | MIIM_TYPE;
    mi.wID = ID_KEYCTX_PASTE;
    InsertMenuItem (hm, 0, FALSE, &mi);
    GetCursorPos (&p);
    TrackPopupMenu (hm, 0, p.x, p.y, 0, dlg, NULL);
    DestroyMenu (hm);
}


/* Reload the key cache if requested. */
static void
do_check_cache (listview_ctrl_t lv, HWND dlg, HWND sb)
{
    gpg_keycache_t cache;

    if (keycache_get_reload ()) {
	keycache_reload (dlg);
	keycache_set_reload (0);
	cache = keycache_get_ctx (1);
	if (!cache)
	    BUG (dlg);
	keylist_reload (lv, cache, KEYLIST_LIST, KEY_SORT_USERID);
	km_update_status_bar (sb, lv);
    }
}


long CALLBACK
separator_wnd_proc (HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam)
{
    static POINT last_mouse_pos;

    if (msg == WM_CREATE)
        SetWindowLong (hwnd, GWL_USERDATA, 
		       (long)(((CREATESTRUCT*)lparam)->lpCreateParams));

    switch (msg) {
    case WM_PAINT:
        PAINTSTRUCT ps;
        RECT rect;
        HPEN hpen;

        GetClientRect (hwnd, &rect);
        BeginPaint (hwnd, &ps);

        /* Background */
        FillRect (ps.hdc, &rect, (HBRUSH)(COLOR_3DFACE+1));

        /* The lines from the light into the dark */
        MoveToEx(ps.hdc, 0,0, NULL);
        if ((hpen = CreatePen (PS_SOLID, 0, GetSysColor(COLOR_3DHILIGHT))) != NULL) {
            SelectObject (ps.hdc, (LPVOID)hpen);
            LineTo (ps.hdc, rect.right, 0);
            DeleteObject (hpen);
        }
	MoveToEx(ps.hdc, 0, 1, NULL);
	if ((hpen = CreatePen (PS_SOLID, 0, GetSysColor(COLOR_3DLIGHT))) != NULL) {
             SelectObject (ps.hdc, (LPVOID)hpen);
             LineTo (ps.hdc, rect.right, rect.bottom);
             DeleteObject (hpen);
         }

        MoveToEx(ps.hdc, 0, rect.bottom-1, NULL);
        if ((hpen = CreatePen (PS_SOLID, 0, GetSysColor(COLOR_3DSHADOW))) != NULL) {
            SelectObject (ps.hdc, (LPVOID)hpen);
            LineTo (ps.hdc, rect.right, rect.bottom-1);
            DeleteObject (hpen);
        }
        MoveToEx(ps.hdc, 0, rect.bottom, NULL);
        if ((hpen = CreatePen(PS_SOLID, 0, GetSysColor(COLOR_3DDKSHADOW))) != NULL) {
            SelectObject (ps.hdc, (LPVOID)hpen);
            LineTo (ps.hdc, rect.right, rect.bottom);
            DeleteObject (hpen);
        }

        EndPaint (hwnd, &ps);
        return 0;

    case WM_LBUTTONDOWN:
        last_mouse_pos.x = LOWORD (lparam);
        last_mouse_pos.y = HIWORD (lparam);
        ClientToScreen (hwnd, &last_mouse_pos);
        SetCapture (hwnd);
        return 0;

    case WM_LBUTTONUP:
        ReleaseCapture ();
        return 0;

    case WM_MOUSEMOVE:
        if (wparam == MK_LBUTTON) {
            struct km_info *kmi;
            POINT p;
            RECT r;

	    kmi = (struct km_info *)GetWindowLong (hwnd, GWL_USERDATA);
            if (kmi == NULL)
                break;

            /* Calculate mouse movement */
            p.x = LOWORD(lparam);
            p.y = HIWORD(lparam);
            ClientToScreen (hwnd, &p);

            GetWindowRect (hwnd, &r);
            r.top += (short)(p.y - last_mouse_pos.y);
            r.bottom += (short)(p.y - last_mouse_pos.y);

            last_mouse_pos.y = p.y;

            /* Apply mouse movement to window. Beware the MoveWindow is relaive 
	       to parent NOT screen */
            MapWindowPoints (NULL, GetParent(hwnd), (POINT*)&r, 2);
            kmi->ypos_sep = r.top;
            kmi->ypercent_sep = 0; /* This forces do_resize_window to use abs. position */
	    do_resize_window (GetParent(hwnd), kmi);
	    return 0;
        }
    }

    return DefWindowProc (hwnd, msg, wparam, lparam);
}


/* Register the separator window with @dlg as the parent window. */
static HWND
regist_sep_wnd (HWND dlg, struct km_info * kmi)
{
    WNDCLASS wndclass;
    HWND h;

    wndclass.style         = CS_DBLCLKS | CS_HREDRAW | CS_VREDRAW;
    wndclass.lpfnWndProc   = separator_wnd_proc;
    wndclass.cbClsExtra    = 0;
    wndclass.cbWndExtra    = sizeof (long);
    wndclass.hInstance     = glob_hinst;
    wndclass.hIcon         = NULL;
    wndclass.hCursor       = LoadCursor (NULL, IDC_SIZENS);
    wndclass.hbrBackground = (HBRUSH) GetStockObject (LTGRAY_BRUSH);
    wndclass.lpszMenuName  = 0;
    wndclass.lpszClassName = WINDOWCLASS_SEPARATOR_CHILD;
    RegisterClass (&wndclass);

    h = CreateWindowEx (WS_EX_WINDOWEDGE, WINDOWCLASS_SEPARATOR_CHILD, NULL,
			WS_CHILD | WS_CLIPCHILDREN | WS_VISIBLE, 
			0, 400, 2000, 5, dlg, (HMENU) 0, glob_hinst, kmi);    
    ShowWindow (h, SW_SHOW);
    return h;
}


#define enable_button(hwnd, cid) \
    SendMessage ((hwnd), TB_ENABLEBUTTON, (cid), MAKELONG (key_selected, 0))


/* Interactive modification of the dialog item which depend if an item
   is selected or not and what kind of item. @lv is the list view with
   the items. */
void
update_ui_items (HWND hwnd, listview_ctrl_t lv)
{
    HWND hwnd_child;
    HMENU hmenu;
    int key_selected = 0;
    int key_has_sec = 0;
    int key_inv = 0;
    int mult_resids[] = {ID_KEYMISC_PROPS, ID_KEYMISC_SIGN, ID_KEYMISC_EDITKEY,
			 ID_KEYMISC_CHECKSIGS, ID_KEYMISC_REVCERT, 0};
    int i;
    
    /* Get some states */
    key_selected = SendMessage (GetDlgItem (hwnd, IDC_KEYMISC_KEYLIST), 
					   LVM_GETSELECTEDCOUNT, 0, 0) 
					   ? TRUE : FALSE;
    key_has_sec = FALSE;
    if (key_selected) {
        i = listview_get_curr_pos (lv);
        key_has_sec = km_check_for_seckey (lv, i, NULL) ? TRUE : FALSE;
	key_inv = km_get_key_status (lv, i) & KM_FLAG_REVOKED ||
		  km_get_key_status (lv, i) & KM_FLAG_EXPIRED;
    }

    /* Enable / disable toolbar buttons */
    hwnd_child = GetDlgItem (hwnd, IDR_WINPT_KMTB);
    enable_button (hwnd_child, ID_KEYMISC_DELETE);
    enable_button (hwnd_child, ID_KEYMISC_PROPS);
    enable_button (hwnd_child, ID_KEYMISC_SIGN);
    enable_button (hwnd_child, ID_KEYMISC_EXPORT);

    /* Enable / disable menu items */
    hmenu = GetMenu (hwnd);
    set_menu_state (hmenu, ID_KEYMISC_EXPORT, key_selected ? MF_ENABLED : MF_GRAYED);
    set_menu_state (hmenu, ID_KEYMISC_EXPORT_PRIVKEY, key_has_sec ? MF_ENABLED : MF_GRAYED);
    set_menu_state (hmenu, ID_KEYMISC_REVCERT, key_has_sec ? MF_ENABLED : MF_GRAYED);
    set_menu_state (hmenu, ID_KEYMISC_DELETE, key_selected ? MF_ENABLED : MF_GRAYED);
    set_menu_state (hmenu, ID_KEYMISC_PROPS, key_selected ? MF_ENABLED : MF_GRAYED);
    set_menu_state (hmenu, ID_KEYMISC_SIGN, key_selected && !key_inv ? MF_ENABLED : MF_GRAYED);
    set_menu_state (hmenu, ID_KEYMISC_EDITKEY, key_selected? MF_ENABLED : MF_GRAYED);
    set_menu_state (hmenu, ID_KEYMISC_CHECKSIGS, key_selected? MF_ENABLED : MF_GRAYED);

    /* Disable some menu items when multiple keys are selected. */
    if (listview_count_items (lv, 1) > 1) {
	for (i=0; mult_resids[i] != 0; i++)
	    set_menu_state (hmenu, mult_resids[i], MF_GRAYED);
    }
}


/* Disable some context menu items when multiple keys are selected. */
static void
popup_multiple (HWND dlg, HMENU hm)
{
    int resids[] = {
	ID_KEYCTX_EDIT,
	ID_KEYCTX_SIGN,
	ID_KEYCTX_REV,
	ID_KEYCTX_ENABLE,
	ID_KEYCTX_DISABLE,
	ID_KEYCTX_ADDKEY,
	ID_KEYCTX_ADDPHOTO,
	ID_KEYCTX_ADDUID,
	ID_KEYCTX_ADDREV,
	ID_KEYCTX_LISTSIGS,
	ID_KEYCTX_MAXTRUST,
	ID_KEYCTX_PROPS,
	ID_KEYCTX_SENDMAIL,
	0};
    int i;
    for (i=0; i < resids[i] != 0; i++)
	set_menu_state (hm, resids[i], MF_GRAYED);
}


/* The items which are shown when GPG is in read-only mode. */
static void
popup_gpg_readonly (HWND dlg, HMENU hm)
{
    int resids[] = {
	ID_KEYCTX_PASTE,
	ID_KEYCTX_EDIT,
	ID_KEYCTX_SIGN,
	ID_KEYCTX_REV,
	ID_KEYCTX_DEL,
	ID_KEYCTX_ENABLE,
	ID_KEYCTX_DISABLE,
	ID_KEYCTX_RECVFROM,
	ID_KEYCTX_ADDKEY,
	ID_KEYCTX_ADDPHOTO,
	ID_KEYCTX_ADDUID,
	ID_KEYCTX_ADDREV,
	0
    };
    int i;

    if (gpg_read_only == 0)
	return;
    for (i=0; resids[i] != 0; i++)
	set_menu_state (hm, resids[i], MF_GRAYED);
}


/* Change the 'Edit' menu based on the current state. */
static void
change_edit_menu (listview_ctrl_t lv, HMENU hm, int id)
{
    int no_sel;

    if (id != 1 /*EDIT_MENU*/)
	return;

    if (!clip_contains_pgpkey ())
	set_menu_state (hm, ID_KEYMISC_PASTE, MF_GRAYED);
    else
	set_menu_state (hm, ID_KEYMISC_PASTE, MF_ENABLED);
    no_sel = listview_get_curr_pos (lv) == -1? 1 : 0;
    set_menu_state (hm, ID_KEYMISC_DELETE2, no_sel? MF_GRAYED: MF_ENABLED);
    set_menu_state (hm, ID_KEYMISC_COPY, no_sel? MF_GRAYED : MF_ENABLED);
}



/* Show limited key menu entries when GPG is in read-only mode. */
static void
change_key_menu (HMENU hm, int id)
{
    int key_resids[] = {
	ID_KEYMISC_SIGN,
	ID_KEYMISC_DELETE,
	ID_KEYMISC_REVCERT,
	ID_KEYMISC_IMPORT_HTTP,
	ID_KEYMISC_IMPORT,
	ID_KEYMISC_REFRESH_KEYS,
	ID_KEYMISC_REBUILD,
	ID_KEYMISC_KEYWIZARD,
	ID_KEYMISC_CARDNEW,
	ID_KEYMISC_NEWKEY,
	0,
    };
    int edit_resids[] = {
	ID_KEYMISC_PASTE,
	ID_KEYMISC_OT,
	ID_KEYMISC_EDITKEY,
	0
    };
    int *resids;
    int i;

    
    if (gpg_read_only == 0)
	return;
    switch (id) {
    case 0: return;
    case 3: resids = key_resids; break;
    case 1: resids = edit_resids;break;
    default:resids = edit_resids; break;
    }

    for (i=0; resids[i] != 0; i++)
	set_menu_state (hm, resids[i], MF_GRAYED);
}


/* Return a user-friendly name for a key derrived from
   name. If @is_secret is 1, a secret key name will be generated. */
static char*
gen_export_filename (const char *keyid, int is_secret)
{
    gpgme_key_t key;
    const char *s;
    char *p;

    if (get_pubkey (keyid, &key))
	return m_strdup (keyid);
    s = key->uids->name;
    if (!s)
	return m_strdup (keyid);
    p = new char[strlen (s) + 8 + 16];
    if (!p)
	BUG (0);
    sprintf (p, "%s%s.asc", s, is_secret? "_sec" : "");
    for (size_t i=0; i < strlen (p); i++) {
	if (p[i] == ' ' || p[i] == ':' || p[i] == '?' || p[i] == '|')
	    p[i] = '_';
    }
    return p;
}


/* Reload a single key in the cache. */
static void
update_key (listview_ctrl_t lv, int pos, const char *keyid, int keypair)
{
    gpgme_key_t key;

    keycache_update (0, keyid);
    if (keypair)
	keycache_update (1, keyid);

    /* because we replaced the key item, we need to update the
       address of the pointer in the ListView control. */
    get_pubkey (keyid, &key);
    keylist_upd_key (lv, pos, key);
}


/* Setup status bar for the main window @dlg. */
static HWND
setup_status_bar (HWND dlg, listview_ctrl_t lv)
{	
    HWND statbar;
    RECT r;
    int partpos[3];
    int i;

    GetClientRect (dlg, &r);
    for (i=1; i <= 3; i++)
	partpos[i-1] = r.right/3*i;
    statbar = CreateStatusWindow (WS_CHILD | WS_VISIBLE | CCS_BOTTOM, 
				  "status bar", dlg, KM_SEPARATOR_ID);

    ShowWindow (statbar, SW_SHOW);
    SendMessage (statbar, SB_SETPARTS, (WPARAM)3, (LPARAM)partpos);

    km_update_default_key_str (statbar);
    km_update_status_bar (statbar, lv);

    return statbar;
}


/* Dialog box procedure for the Key Manager. */
BOOL CALLBACK
keymanager_dlg_proc (HWND dlg, UINT msg, WPARAM wparam, LPARAM lparam)
{
    struct km_info *kmi;
    static HMENU menu = NULL;
    static int refresh_keys = 0;    
    INITCOMMONCONTROLSEX icex;
    HWND kl;
    HMENU hm;
    gpg_keycache_t c;
    gpgme_key_t key;
    struct genkey_s genkey;
    struct winpt_key_s k = {0};
    struct URL_ctx_s *url;
    refresh_cache_s rcs = {0};
    char keyid[48], uid[128], type[32], *name;
    const char *t, *host;
    u16 port = 0;
    int l_idx = 0, i=0, rc;

    if ((msg != WM_INITDIALOG) 
	&& ((kmi = (struct km_info*)GetWindowLong (dlg, GWL_USERDATA)) == NULL))
        return FALSE;

    switch (msg) {
    case WM_INITDIALOG:
        kmi = new struct km_info;
        memset (kmi, 0, sizeof (struct km_info));
	kmi->lv_idx = -1;
        icex.dwSize = sizeof (INITCOMMONCONTROLSEX);
        icex.dwICC  = ICC_BAR_CLASSES;
        InitCommonControlsEx (&icex);
	kmi->hwnd_sep = regist_sep_wnd (dlg, kmi);
        imagelist_load (dlg);
	SetWindowText (dlg, _("Key Manager"));
        menu = LoadMenu (glob_hinst, (LPCSTR)IDR_WINPT_KEYMISC);
	set_menu_text_bypos (menu, 0, _("File"));
	set_menu_text_bypos (menu, 1, _("Edit"));
	set_menu_text_bypos (menu, 2, _("View"));
	set_menu_text_bypos (menu, 3, _("Key"));
	set_menu_text_bypos (menu, 4, _("Groups"));
	set_menu_text (menu, ID_KEYMISC_EDITKEY, _("Edit"));
	set_menu_text (menu, ID_KEYMISC_MAIL, _("Send Mail..."));
	set_menu_text (menu, ID_KEYMISC_OT, _("Ownertrust")); /* XXX */
	set_menu_text (menu, ID_KEYMISC_COPY, _("&Copy\tCtrl+C"));
	set_menu_text (menu, ID_KEYMISC_PASTE, _("&Paste\tCtrl+V"));
	set_menu_text (menu, ID_KEYMISC_FIND, _("Search...\tCtrl+F"));
	set_menu_text (menu, ID_KEYMISC_SELALL, _("Select All\tCtrl+A"));
        set_menu_text (menu, ID_KEYMISC_QUIT, _("&Quit"));
        set_menu_text (menu, ID_KEYMISC_UID, _("User ID"));
        set_menu_text (menu, ID_KEYMISC_NEWKEY, _("&Expert"));
	set_menu_text (menu, ID_KEYMISC_KEYWIZARD, _("&Normal"));
        set_menu_text (menu, ID_KEYMISC_EDIT, _("Edit"));
        set_menu_text (menu, ID_KEYMISC_SIGN, _("&Sign"));
        set_menu_text (menu, ID_KEYMISC_DELETE, _("&Delete"));
	set_menu_text (menu, ID_KEYMISC_DELETE2, _("&Delete"));
        set_menu_text (menu, ID_KEYMISC_REVCERT, _("&Revoke"));
        set_menu_text (menu, ID_KEYMISC_CHECKSIGS, _("&List Signatures"));
        set_menu_text (menu, ID_KEYMISC_TRUSTPATH, _("List Trust Path"));
	set_menu_text (menu, ID_KEYMISC_EXPORT, _("&Export..."));
	set_menu_text (menu, ID_KEYMISC_IMPORT, _("&Import..."));
        set_menu_text (menu, ID_KEYMISC_PROPS, _("&Properties"));
        set_menu_text (menu, ID_KEYMISC_GPGOPT, _("Options"));
        set_menu_text (menu, ID_KEYMISC_GPGPREFS, _("Preferences"));
        set_menu_text (menu, ID_KEYMISC_SENDRECV, _("Keyserver") );
	set_menu_text (menu, ID_KEYMISC_EXPORT_PRIVKEY, _("E&xport Secret Key"));
        set_menu_text (menu, ID_KEYMISC_RECACHE, _("Re&load Key Cache"));
        set_menu_text (menu, ID_KEYMISC_REBUILD, _("R&everify Signatures"));
	set_menu_text (menu, ID_KEYMISC_REFRESH_KEYS, _("Refresh &Keys (Keyserver)"));
	set_menu_text (menu, ID_KEYMISC_INFO, _("Info") );
	set_menu_text (menu, ID_KEYMISC_HELP, _("&Help"));
	
	/* XXX: implement help code. */
	set_menu_state (menu, ID_KEYMISC_HELP, MF_GRAYED);
	
        SetMenu (dlg, menu);
        if (keyring_check_last_access ())
            keycache_set_reload (1);
        if (keycache_get_reload ())
            keycache_reload (dlg);
        c = keycache_get_ctx (KEYCACHE_PUB);
        if (!c)
            BUG (NULL);
        kmi->keylist_sortby = KEY_SORT_USERID;
        Header_SetImageList(ListView_GetHeader (GetDlgItem (dlg, IDC_KEYMISC_KEYLIST)),
			    glob_imagelist);
        kmi->lv = keylist_load (GetDlgItem (dlg, IDC_KEYMISC_KEYLIST), c, 
				NULL, KEYLIST_LIST, kmi->keylist_sortby);
        /* init subclassing for the listview */
        kl = GetDlgItem (dlg, IDC_KEYMISC_KEYLIST);
        keylist_proc.dlg = dlg;
        keylist_proc.current = (WNDPROC)keylist_subclass_proc;
        keylist_proc.old = (WNDPROC)GetWindowLong( kl, GWL_WNDPROC );
        if (keylist_proc.old) {
            if( !SetWindowLong (kl, GWL_WNDPROC, (LONG)keylist_proc.current)) {
                msg_box (dlg, _("Could not set keylist window procedure."), 
			 _("Key Manager"), MB_ERR);
		BUG (NULL);
	    }
        }
	#if 0
        km_groups_new (&gc, GetDlgItem (dlg, IDC_KEYMISC_GROUP));
        km_groups_load (gc);
	#endif
	kmi->statbar = setup_status_bar (dlg, kmi->lv);
        SetWindowLong (dlg, GWL_USERDATA, (LONG)kmi);
        kmi->toolbar = load_toolbar (dlg, kmi);

	do_center_window (dlg, kmi);
        do_resize_window (dlg, kmi);
        update_ui_items (dlg, kmi->lv);

	SetDlgItemText (dlg, IDC_KEYMISC_GTEXT, _("Groups"));
	SetClassLong (dlg, GCL_HICON, (LONG)LoadIcon (glob_hinst, 
		      (LPCTSTR)IDI_WINPT));
	SetForegroundWindow (dlg);
        return TRUE;
        
    case WM_DESTROY:
	if (kmi->lv) {
            keylist_delete (kmi->lv);
            kmi->lv = NULL;       
	}
	imagelist_destroy ();

	ltoa (kmi->pos_x, type, 10);
	set_reg_entry (HKEY_CURRENT_USER, "Software\\WinPT", "KM_Pos_X", type);
	ltoa (kmi->pos_y, type, 10);
	set_reg_entry (HKEY_CURRENT_USER, "Software\\WinPT", "KM_Pos_Y", type);
	/* Remove runtime information. This should be the last action taken here. */
	delete kmi; kmi = NULL;
	SetWindowLong (dlg, GWL_USERDATA, 0);
	keycache_set_reload (refresh_keys);
	return FALSE;

    case WM_MOVE:
        RECT r;
        GetWindowRect (dlg, &r);
	kmi->pos_x = r.left;
	kmi->pos_y = r.top;	
	break;
        

    case WM_RBUTTONUP:
	do_create_minpopup (dlg);
	break;

    case WM_NOTIFY:		
        NMHDR *notify;
        POINT p;
        HMENU popup;
        
        notify = (NMHDR *)lparam;
	if (!notify)
	    break;
	switch (notify->code) {
        case TBN_QUERYDELETE:
	    SetWindowLong (dlg, DWL_MSGRESULT, TRUE);
            return TRUE;
        
        case TBN_QUERYINSERT:
	    SetWindowLong (dlg, DWL_MSGRESULT, TRUE);
            return TRUE;

        case TBN_GETBUTTONINFO:
            LPTBNOTIFY lpTbNotify;
            lpTbNotify = (LPTBNOTIFY)lparam;
            if (lpTbNotify->iItem < (sizeof(myb) / sizeof(mybuttons))) {
                lpTbNotify->tbButton.iBitmap = imagelist_getindex(myb[lpTbNotify->iItem].icon);
                lpTbNotify->tbButton.idCommand = myb[lpTbNotify->iItem].command;
                lpTbNotify->tbButton.fsState = TBSTATE_ENABLED;
                lpTbNotify->tbButton.fsStyle = TBSTYLE_BUTTON;
                lpTbNotify->tbButton.iString = 0;
                strncpy (lpTbNotify->pszText, myb[lpTbNotify->iItem].text, lpTbNotify->cchText);
                SetWindowLong(dlg, DWL_MSGRESULT, TRUE);
            }
            return TRUE;

        case TBN_RESET: /* Restore last saved toolbar status */
            TBSAVEPARAMS tbsp;

            tbsp.hkr = HKEY_CURRENT_USER;
            tbsp.pszSubKey = "Software\\WinPT";
            tbsp.pszValueName = "KM_toolbar";
            SendMessage (notify->hwndFrom, TB_SAVERESTORE, FALSE, (LPARAM)&tbsp); 
            break;

        case TBN_BEGINADJUST: /* User is about to change the toolbar. Save it */
            tbsp.hkr = HKEY_CURRENT_USER;
            tbsp.pszSubKey = "Software\\WinPT";
            tbsp.pszValueName = "KM_toolbar";
            SendMessage (notify->hwndFrom, TB_SAVERESTORE, TRUE, (LPARAM)&tbsp); 
            break;

        case TBN_ENDADJUST: /* User has finished customizing the toolbar. Save it. */
            tbsp.hkr = HKEY_CURRENT_USER;
            tbsp.pszSubKey = "Software\\WinPT";
            tbsp.pszValueName = "KM_toolbar";
            SendMessage (notify->hwndFrom, TB_SAVERESTORE, TRUE, (LPARAM)&tbsp); 
            break;

        case TTN_GETDISPINFO:
            LPTOOLTIPTEXT lpttt;
            lpttt = (LPTOOLTIPTEXT)lparam;

            lpttt->hinst = NULL;
            switch (lpttt->hdr.idFrom) {
            case ID_KEYMISC_DELETE:
                lpttt->lpszText = (char *)_("Delete key from keyring"); 
		break;

            case ID_KEYMISC_PROPS:
                lpttt->lpszText = (char *)_("Show key properties"); 
		break;

            case ID_KEYMISC_SIGN:
                lpttt->lpszText = (char *)_("Sign key"); 
		break;

            case ID_KEYMISC_IMPORT:
                lpttt->lpszText = (char *)_("Import key to keyring");
		break;

            case ID_KEYMISC_EXPORT:
                lpttt->lpszText = (char *)_("Export key to a file"); 
		break;
            }
            return TRUE;
            
        case LVN_ITEMCHANGED:
            if (((LPNMLISTVIEW)lparam)->uNewState) { /* item selected? */
		kmi->lv_idx = listview_get_curr_pos (kmi->lv);
                update_ui_items (dlg, kmi->lv);
                return TRUE;
            }
            break;

	case NM_RCLICK:
            if (notify->idFrom == IDC_KEYMISC_KEYLIST) {
		l_idx =listview_get_curr_pos (kmi->lv);
		if (l_idx == -1)
		    return TRUE; /* Popup only when a item was selected */
		do_check_cache (kmi->lv, dlg, kmi->statbar);
		GetCursorPos (&p);
		hm = LoadMenu (glob_hinst, MAKEINTRESOURCE (IDR_WINPT_KEYMISC_CTX));
		popup = GetSubMenu (hm, 0);

		set_menu_text (popup, ID_KEYCTX_UID_COPY, _("Copy User ID to Clipboard"));
		set_menu_text (popup, ID_KEYCTX_KEYID_COPY, _("Copy Key ID to Clipboard"));
		set_menu_text (popup, ID_KEYCTX_FPR_COPY, _("Copy Fingerprint to Clipboard"));
		set_menu_text (popup, ID_KEYCTX_KINFO_COPY, _("Copy Key Info to Clipboard"));
		set_menu_text (popup, ID_KEYCTX_COPY, _("Copy Key to Clipboard"));
		set_menu_text (popup, ID_KEYCTX_PASTE, _("Paste Key from Clipboard"));
		set_menu_text (popup, ID_KEYCTX_RECVFROM, _("Refresh from Keyserver"));
		set_menu_text (popup, ID_KEYCTX_MAXTRUST, _("Set Implicit &Trust"));
		set_menu_text (popup, ID_KEYCTX_LISTSIGS, _("&List Signatures"));
		set_menu_text (popup, ID_KEYCTX_PROPS, _("&Key Properties"));
		set_menu_text (popup, ID_KEYCTX_EDIT, _("Key Edit"));
		set_menu_text (popup, ID_KEYCTX_DEL, _("&Delete"));
		set_menu_text (popup, ID_KEYCTX_REV, _("&Revoke"));
		set_menu_text (popup, ID_KEYCTX_SIGN, _("&Sign"));
		set_menu_text (popup, ID_KEYCTX_ENABLE, _("&Enable"));
		set_menu_text (popup, ID_KEYCTX_DISABLE, _("&Disable"));
		set_menu_text (popup, ID_KEYCTX_RECVFROM, _("Re&fresh from Keyserver"));
		set_menu_text (popup, ID_KEYCTX_SETPREFKS, _("Set preferred Keyserver URL"));
		set_menu_text (popup, ID_KEYCTX_SENDMAIL, _("Send Key to Mail Recipient"));
		set_menu_text (popup, ID_KEYCTX_SETDEFKEY, _("Set as Default Key"));

		set_menu_text (popup, ID_KEYCTX_ADDKEY, _("Key..."));
		set_menu_text (popup, ID_KEYCTX_ADDUID, _("User ID..."));
		set_menu_text (popup, ID_KEYCTX_ADDPHOTO, _("Photo ID..."));
		set_menu_text (popup, ID_KEYCTX_ADDREV, _("Revoker..."));

		/* change popup texts */
		set_menu_text_bypos (popup, 0, _("Key Attributes"));
		set_menu_text_bypos (popup, 6, _("Add"));
		set_menu_text_bypos (popup, 19, _("Send to Keyserver"));

		if (km_check_for_seckey (kmi->lv, l_idx, &i))
		    set_menu_state (popup, ID_KEYCTX_SETDEFKEY, MF_ENABLED);
		if (i == 0)
		    set_menu_state (popup, ID_KEYCTX_MAXTRUST, MF_ENABLED);
		if (!km_check_for_seckey (kmi->lv, l_idx, NULL)) {
		    set_menu_state (popup, ID_KEYCTX_REV, MF_DISABLED|MF_GRAYED);
		    set_menu_state (popup, ID_KEYCTX_ADDKEY, MF_DISABLED|MF_GRAYED);
		    set_menu_state (popup, ID_KEYCTX_ADDUID, MF_DISABLED|MF_GRAYED);
		    set_menu_state (popup, ID_KEYCTX_ADDREV, MF_DISABLED|MF_GRAYED);
		    set_menu_state (popup, ID_KEYCTX_ADDPHOTO, MF_DISABLED|MF_GRAYED );
		    set_menu_state (popup, ID_KEYCTX_SETPREFKS, MF_DISABLED|MF_GRAYED);
		}
		else if (km_check_for_seckey (kmi->lv, l_idx, NULL) &&
		         km_key_is_v3 (kmi->lv, l_idx)) {
		    /* PGP 2 keys, version 3 have no no support for photo-id's,
		       designated revokers and secondary keys. */
		    set_menu_state (popup, ID_KEYCTX_ADDKEY, MF_DISABLED|MF_GRAYED);
		    set_menu_state (popup, ID_KEYCTX_ADDREV, MF_DISABLED|MF_GRAYED);
		    set_menu_state (popup, ID_KEYCTX_ADDPHOTO, MF_DISABLED|MF_GRAYED);
		}
		if (km_get_key_status( kmi->lv, l_idx ) & KM_FLAG_DISABLED)
		    set_menu_state (popup, ID_KEYCTX_DISABLE, MF_DISABLED|MF_GRAYED);
		else
		    set_menu_state (popup, ID_KEYCTX_ENABLE, MF_DISABLED|MF_GRAYED);
		if (km_get_key_status (kmi->lv, l_idx) & KM_FLAG_REVOKED ||
		    km_get_key_status (kmi->lv, l_idx) & KM_FLAG_EXPIRED)
		    set_menu_state (popup, ID_KEYCTX_SIGN, MF_DISABLED|MF_GRAYED);
		if (!clip_contains_pgpkey ())
		    set_menu_state (popup, ID_KEYCTX_PASTE, MF_DISABLED|MF_GRAYED);
		if (mapi_init ())
		    set_menu_state (popup, ID_KEYCTX_SENDMAIL, MF_DISABLED|MF_GRAYED);
		/* Override 'Default Keyserver' with the actual name. */
		host = kserver_get_hostname (0, -1, &port);
		set_menu_text (popup, ID_KEYCTX_KS_DEFAULT, host);
		popup_gpg_readonly (dlg, popup);
		if (listview_count_items (kmi->lv, 1) > 1)
		    popup_multiple (dlg, popup);
		TrackPopupMenu (popup, TPM_RIGHTALIGN, p.x, p.y, 0, dlg, NULL);
		DestroyMenu (popup);
		DestroyMenu (hm);
		return TRUE;
	    }
	    #if 0 /* XXX */
	    if( notify->idFrom == IDC_KEYMISC_GROUP ) {
		HWND tree = GetDlgItem( dlg, IDC_KEYMISC_GROUP );
		if( TreeView_GetSelection( tree ) ) {
		    GetCursorPos( &p );
		    hm = LoadMenu( glob_hinst, MAKEINTRESOURCE(IDR_WINPT_GROUP_CTX) );
		    popup = GetSubMenu (hm, 0);
		    if (km_index == -1) /* XXX */
			set_menu_state( popup, ID_GROUP_PASTE, MF_DISABLED|MF_GRAYED );
		    set_menu_text( popup, ID_GROUP_PASTE, _("Paste into this group") );
		    set_menu_text( popup, ID_GROUP_DELETE, _("Delete") );
		    TrackPopupMenu( popup, TPM_RIGHTALIGN, p.x, p.y, 0, dlg, NULL );
		    DestroyMenu( popup );
		    DestroyMenu( hm );
		    return TRUE;
		}
	    }
	    #endif
	    break;

        case LVN_COLUMNCLICK:
	    if (notify->idFrom == IDC_KEYMISC_KEYLIST) {
                NMLISTVIEW *nft = (LPNMLISTVIEW) lparam;
                int sortby = 0;
                switch (nft->iSubItem) {
                case 0:  sortby = KEY_SORT_USERID; break;
                case 1:  sortby = KEY_SORT_KEYID; break;
                case 2:  sortby = KEY_SORT_IS_SECRET; break;
                case 3:  sortby = KEY_SORT_LEN; break;
                case 5:  sortby = KEY_SORT_VALIDITY; break;
		case 6:  sortby = KEY_SORT_OTRUST; break;
                case 7:  sortby = KEY_SORT_CREATED; break;
		case 8:  sortby = KEY_SORT_ALGO; break;
                default: return TRUE; //sortby = KEY_SORT_USERID; break;
                }

                if ((kmi->keylist_sortby & ~KEYLIST_SORT_DESC) == sortby)
                    kmi->keylist_sortby ^= KEYLIST_SORT_DESC;
                else
                    kmi->keylist_sortby = sortby;
                keylist_sort (kmi->lv, kmi->keylist_sortby);
                return TRUE;
            }
            break;
        }
        break;

    case WM_WINDOWPOSCHANGING:
        if (((WINDOWPOS*)lparam)->cx < 400)
            ((WINDOWPOS*)lparam)->cx = 400;
        if (((WINDOWPOS*)lparam)->cy < 200)
            ((WINDOWPOS*)lparam)->cy = 200;
         return TRUE;
        
    case WM_SIZE:
        do_resize_window (dlg, kmi);
        return TRUE;
        
    case WM_SYSCOMMAND:
        if (LOWORD (wparam) == SC_CLOSE)
            EndDialog (dlg, TRUE);
        return FALSE;
        
    case WM_MENUSELECT:
	change_edit_menu (kmi->lv, (HMENU)lparam, LOWORD (wparam));
	change_key_menu ((HMENU)lparam, LOWORD (wparam));
	break;

    case WM_INITMENUPOPUP:
	if ((UINT)LOWORD (lparam) == 3) {
	    HMENU h = (HMENU)wparam;
	    set_menu_text_bypos (h, 0, _("New"));
	}
	return FALSE;

    case WM_COMMAND:
	/* Allow at least 'Exit' in such a case. */
        if (gnupg_access_keyring (1) && LOWORD (wparam) != ID_KEYMISC_QUIT) {
            msg_box (dlg, _("Could not access public keyring"), 
		     _("Key Manager"), MB_ERR);
            return FALSE;
        }
	do_check_cache (kmi->lv, dlg, kmi->statbar);
        switch( LOWORD (wparam)) {
        case ID_KEYMISC_QUIT:
            EndDialog (dlg, TRUE);
            return TRUE;
	    
	case ID_KEYMISC_MAIL:
	    /* XXX
	    DialogBoxParam (glob_hinst, (LPCTSTR)IDD_WINPT_MAIL, GetDesktopWindow (),
			    winpt_mail_proc, NULL);*/
	    break;
        
	case ID_KEYMISC_FIND:
	    km_find_key (dlg, kmi->lv);
	    break;

        case ID_KEYMISC_DELETE:
	case ID_KEYMISC_DELETE2:
            if (!km_delete_keys (kmi->lv, dlg))
		km_update_status_bar (kmi->statbar, kmi->lv);
            return TRUE;
            
        case ID_KEYMISC_SIGN:
	    if (kmi->lv_idx == -1) {
                msg_box (dlg, _("Please select a key."),  
		         _("Key Manager"), MB_ERR);
                return TRUE;;
            }
            if (km_check_key_status (kmi->lv, kmi->lv_idx))
                return TRUE;
	    key = (gpgme_key_t)listview_get_item2 (kmi->lv, kmi->lv_idx);
	    listview_get_item_text (kmi->lv, kmi->lv_idx, 1, keyid, DIM (keyid)-1);
	    memset (&k, 0, sizeof (k));
	    k.ctx = key;
	    k.keyid = keyid;
	    dialog_box_param (glob_hinst, (LPCSTR)IDD_WINPT_KEYSIGN, dlg,
                              keysign_dlg_proc, (LPARAM)&k, 
                              _("Key Signing"), IDS_WINPT_KEYSIGN);
	    if (k.update)
		update_key (kmi->lv, kmi->lv_idx, k.keyid, 0);
            return TRUE;
            
        case ID_KEYMISC_REVCERT:
            if (kmi->lv_idx == -1) {
                msg_box (dlg, _("Please select a key."), _("Key Manager"), MB_ERR);
                return TRUE;
            }
            listview_get_item_text (kmi->lv, kmi->lv_idx, 0, uid, sizeof uid-1);
            listview_get_item_text (kmi->lv, kmi->lv_idx, 1, keyid, sizeof keyid-1);
            if ( !km_check_for_seckey( kmi->lv, kmi->lv_idx, NULL ) ) {
                msg_box (dlg, _("There is no secret key available!"), 
			_("Key Manager"), MB_ERR);
                return TRUE;
            }
            
            {
                char state[64];
                listview_get_item_text (kmi->lv, kmi->lv_idx, 5, 
					state, sizeof (state) -1);
                if( strchr (state, 'R' )) {
                    msg_box (dlg, _("Key already revoked!"), 
			     _("Key Manager"), MB_INFO);
                    return TRUE;
                }
            }
            
	    memset (&k, 0, sizeof (k));
            k.key_pair = 1;
            k.keyid = keyid;
	    k.is_protected = km_check_if_protected (kmi->lv, kmi->lv_idx);
            dialog_box_param(glob_hinst, (LPCSTR)IDD_WINPT_KEYREVOKE, dlg,
                             key_revoke_dlg_proc, (LPARAM)&k,
                             _("Key Revocation"), IDS_WINPT_KEYREVOKE);
            return TRUE;
            
        case ID_KEYMISC_TRUSTPATH:
	    if (kmi->lv_idx == -1) {
                msg_box (dlg, _("Please select a key."), _("Key Manager"), MB_ERR);
                return TRUE;
            }
            listview_get_item_text( kmi->lv, kmi->lv_idx, 0, uid, sizeof uid -1 );
            listview_get_item_text( kmi->lv, kmi->lv_idx, 1, keyid, sizeof keyid -1 );
            if( km_check_for_seckey( kmi->lv, kmi->lv_idx, NULL ) ) {
                msg_box( dlg, _("It does not make any sense with a key pair!"), _("Key Manager"), MB_OK );
                return TRUE;
            }
	    memset (&k, 0, sizeof (k));
            k.keyid = keyid;
            k.uid = uid;
            dialog_box_param( glob_hinst, (LPCSTR)IDD_WINPT_KEYTRUST, dlg, 
                              keytrust_dlg_proc, (LPARAM)&k, 
                              _("List Trust Path"), IDS_WINPT_KEYTRUST );
            return TRUE;
            
        case ID_KEYMISC_CHECKSIGS:	    
            if (kmi->lv_idx == -1) {
                msg_box (dlg, _("Please select a key."), _("Key Manager"), MB_ERR);
                return TRUE;
            }
            listview_get_item_text (kmi->lv, kmi->lv_idx, 0, uid, DIM (uid)-1);
            listview_get_item_text (kmi->lv, kmi->lv_idx, 1, keyid, DIM (keyid)-1);
	    memset (&k, 0, sizeof (k));
            k.keyid = keyid;
            k.uid = uid;
	    k.ctx = (gpgme_key_t)listview_get_item2 (kmi->lv, kmi->lv_idx);
            dialog_box_param (glob_hinst, (LPCSTR)IDD_WINPT_KEYSIG, dlg, 
                              keysig_dlg_proc, (LPARAM)&k,
                              _("Key Signature List" ), IDS_WINPT_KEYSIG);
            return TRUE;
            
        case ID_KEYMISC_PROPS:	    
            if (kmi->lv_idx == -1) {
                msg_box( dlg, _("Please select a key."), _("Key Manager"), MB_ERR );
                return TRUE;
            }
            listview_get_item_text (kmi->lv, kmi->lv_idx, 1, keyid, DIM (keyid)-1);
            listview_get_item_text (kmi->lv, kmi->lv_idx, 2, type, DIM (type)-1);
	    memset (&k, 0, sizeof (k));
            k.key_pair = 0;
            k.keyid = keyid;
            if( !strcmp( type, "pub/sec" ) || !strcmp( type, "pub/crd" ) )
                k.key_pair = 1;
	    k.callback.ctl = kmi->lv;
	    k.callback.idx = kmi->lv_idx;
	    k.is_v3 = km_key_is_v3 (kmi->lv, kmi->lv_idx);
            dialog_box_param( glob_hinst, (LPCSTR)IDD_WINPT_KEYPROPS, dlg,
                              keyprops_dlg_proc, (LPARAM)&k,
                              _("Key Properties"), IDS_WINPT_KEYPROPS );
	    if (k.callback.new_val != 0) {
		t = get_key_trust_str (k.callback.new_val);
		listview_add_sub_item (kmi->lv, kmi->lv_idx, 6, t);
	    }
            return TRUE;
            
        case ID_KEYMISC_RECACHE:
            /* If there is already a reload request, don't bother the user with a message. */
            if (keycache_get_reload() == 1)
                l_idx = IDYES;
            else {
                char inf[256];
                _snprintf (inf, sizeof (inf) -1,
                           _("This is only useful when the keyring has been "
                             "modified (sign a key...).\n"
                             "Do you really want to reload the keycache?"));
                l_idx = msg_box (dlg, inf, _("Key Manager"), MB_YESNO);
            }
            if (l_idx == IDYES) {
                rcs.kr_reload = rcs.kr_update = 1;
                rcs.tr_update = 0;
                DialogBoxParam (glob_hinst, (LPCSTR)IDD_WINPT_KEYCACHE, dlg,
                                keycache_dlg_proc, (LPARAM)&rcs);
                c = keycache_get_ctx (1);
		if (!c)
		    BUG (dlg);
                keylist_reload (kmi->lv, c, KEYLIST_LIST, KEY_SORT_USERID);
		refresh_keys = 0;
            }
            return TRUE;
            
        case ID_KEYMISC_REBUILD:
	    name = NULL;
	    gpg_rebuild_cache (&name);
	    if (name != NULL) {
		char *line = strchr (name, '\n');
		show_msg (dlg, 2000, line? name + (line-name)+1 : name);
		safe_free (name);
	    }
            return TRUE;
            
        case ID_KEYMISC_NEWKEY:
	    memset (&genkey, 0, sizeof (genkey));
            dialog_box_param (glob_hinst, (LPCSTR)IDD_WINPT_KEYGEN, dlg,
                              keygen_dlg_proc, (LPARAM)&genkey, _("Key Generation"),
			      IDS_WINPT_KEYGEN);
	    if (genkey.newkey != NULL) {
		keylist_add_key (kmi->lv, KEYLIST_LIST, genkey.newkey);
		keylist_sort (kmi->lv, KEY_SORT_USERID);
	    }
            return TRUE;

	case ID_KEYMISC_CARDNEW:
	    if (!scard_support) {
		msg_box (dlg, _("Smart Card support is not available."), 
			 _("Key Manager"), MB_INFO);
		return TRUE;
	    }
	    dialog_box_param (glob_hinst, (LPCSTR)IDD_WINPT_CARD_KEYGEN, dlg,
			      card_keygen_dlg_proc, 0, _("Card Key Generation"),
			      IDS_WINPT_CARD_KEYGEN);
	    /* XXX: use new code */
	    if (keycache_get_reload ())
		send_cmd_id (dlg, ID_KEYMISC_RECACHE);
	    return TRUE;

	case ID_KEYMISC_KEYWIZARD:
	    memset (&genkey, 0, sizeof (genkey));
	    dialog_box_param (glob_hinst, (LPCSTR)IDD_WINPT_KEYWIZARD, dlg,
			      keygen_wizard_dlg_proc, (LPARAM)&genkey, 
			      _("Key Generation Wizard"),
			      IDS_WINPT_KEYWIZARD);
	    if (genkey.newkey != NULL) {
		keylist_add_key (kmi->lv, KEYLIST_LIST, genkey.newkey);
		keylist_sort (kmi->lv, KEY_SORT_USERID);
	    }
	    return TRUE;
            
        case ID_KEYMISC_SENDRECV:
	    memset (&genkey, 0, sizeof (genkey));
            dialog_box_param (glob_hinst, (LPCSTR)IDD_WINPT_KEYSERVER, dlg,
                              keyserver_dlg_proc, (LPARAM)&genkey, _("Keyserver Access"),
			      IDS_WINPT_KEYSERVER);
	    if (genkey.newkey != NULL) {
		keylist_add_key (kmi->lv, KEYLIST_LIST, genkey.newkey);
		keylist_sort (kmi->lv, KEY_SORT_USERID);
	    }
            return TRUE;
            
        case ID_KEYMISC_GPGPREFS:
            dialog_box_param( glob_hinst, (LPCSTR)IDD_WINPT_GPGPREFS, dlg,
                              gpgprefs_dlg_proc, 0, _("GnuPG Preferences"), 
			      IDS_WINPT_GPGPREFS );
            return TRUE;
            
        case ID_KEYMISC_GPGOPT:
            dialog_box_param( glob_hinst, (LPCSTR)IDD_WINPT_GPGOPT, dlg,
                              gpgopt_dlg_proc, 0, _("GnuPG Options" ), 
			      IDS_WINPT_GPGOPT );
            return TRUE;
            
        case ID_KEYMISC_IMPORT:
            t = get_fileopen_dlg (dlg, _("Choose Name of the Key File"), NULL, NULL);
            if (t)
                km_file_import (dlg, t);
            return TRUE;

	case ID_KEYMISC_IMPORT_HTTP:
	    url = (struct URL_ctx_s*)get_http_file_dlg (dlg);
	    if (url->cancel == 0)
		km_http_import (dlg, url->url);
	    delete url; url=NULL;
	    break;
            
        case ID_KEYMISC_EXPORT:
            if (kmi->lv_idx == -1) {
                msg_box( dlg, _("Please select a key."), _("Key Manager"), MB_ERR );
                return TRUE;
            }
	    if (listview_count_items (kmi->lv, 1) > 1)
		name = m_strdup ("Exported_GPG_Keys.asc");
	    else {
		listview_get_item_text (kmi->lv, kmi->lv_idx, 1, keyid, DIM (keyid)-1);
		name = gen_export_filename (keyid, 0);
	    }
            t = get_filesave_dlg (dlg, _("Choose Name for Key File"), NULL, name);
	    free_if_alloc (name);
            if (t == NULL)
                return TRUE;
            km_file_export (dlg, kmi->lv, t);
            return TRUE;
            
        case ID_KEYMISC_EXPORT_PRIVKEY:
            if (kmi->lv_idx == -1) {
                msg_box (dlg, _("Please select a key."), _("Key Manager"), MB_ERR);
                return TRUE;
            }
            if( !km_check_for_seckey( kmi->lv, kmi->lv_idx, NULL ) ) {
                msg_box (dlg, _("There is no corresponding secret key for this key."),
                         _("Key Manager"), MB_ERR);
                return TRUE;
            }
            if (listview_count_items (kmi->lv, 1) > 1) {
                msg_box (dlg, _("You can only export one secret key."), 
			 _("Key Manager"), MB_ERR);
                return TRUE;
            }
            i = msg_box (dlg, 
                          _("This operation will export your *SECRET* key!\n\n"
                            "Never send this key to ANYONE, it should be available\n"
                            "ONLY on your machine and you may use this function\n"
                            "to copy the key to a safe place.\n\n"
                            "Do you really want to export the key?"),
                          _("WARNING"), MB_INFO|MB_YESNO);
            if( i == IDYES ) {
		listview_get_item_text( kmi->lv, kmi->lv_idx, 1, keyid, sizeof (keyid)-8 );
		name = gen_export_filename (keyid, 1);
                t = get_filesave_dlg (dlg, _("Choose Name for Key File"), NULL, name);
                if (t != NULL)		
		    km_privkey_export (dlg, kmi->lv, t);
            }
            return TRUE;

	case ID_KEYMISC_INFO:
	    dialog_box_param (glob_hinst, (LPCSTR)IDD_WINPT_ABOUT, glob_hwnd,
			      about_winpt_dlg_proc, 0, _("About WinPT"),
			      IDS_WINPT_ABOUT);
	    break;

	case ID_KEYMISC_HELP:
	    ShellExecute (dlg, "open", "winpt.chm", NULL, NULL, SW_SHOW);
	    break;

	case ID_KEYMISC_OT:
	    dialog_box_param( glob_hinst, (LPCTSTR)IDD_WINPT_OWNERTRUST, glob_hwnd,
			      ownertrust_dlg_proc, 0,
			      _("Ownertrust"), IDS_WINPT_OWNERTRUST );
	    break;

	case ID_KEYMISC_EDITKEY:	   
	    if (kmi->lv_idx == -1)
		break;
	    listview_get_item_text (kmi->lv, kmi->lv_idx, 1, keyid, sizeof (keyid)-1);
	    /* XXX: pub/crd = secret key does not work */
	    memset (&k, 0, sizeof (k));
	    k.is_protected = km_check_if_protected (kmi->lv, kmi->lv_idx);
	    k.key_pair = km_check_for_seckey (kmi->lv, kmi->lv_idx, NULL);
	    k.keyid = keyid;
	    k.is_v3 = km_key_is_v3 (kmi->lv, kmi->lv_idx);
	    k.flags = km_get_key_status (kmi->lv, kmi->lv_idx);
	    dialog_box_param (glob_hinst, (LPCTSTR)IDD_WINPT_KEYEDIT, dlg,
			      keyedit_main_dlg_proc, (LPARAM)&k,
			      _("Key Edit"), IDS_KEYCTX_EDIT);
	    if (k.update)
		update_key (kmi->lv,  kmi->lv_idx, keyid, 1);
	    break;
            
        case ID_KEYMISC_COPY:
            km_clip_export (dlg, kmi->lv);
            break;
            
	case ID_KEYMISC_SELALL:
	    listview_select_all (kmi->lv);
	    break;

        case ID_KEYMISC_PASTE:
            km_clip_import (dlg);
            break;
            
	case ID_KEYCTX_SETPREFKS:
	    listview_get_item_text (kmi->lv, kmi->lv_idx, 1, keyid, DIM(keyid)-1);
	    memset (&k, 0, sizeof (k));
	    k.keyid = keyid;
	    keyedit_set_pref_keyserver (&k, dlg);
	    break;

        case ID_KEYMISC_REFRESH_KEYS:
	    if (listview_count_items (kmi->lv, 1) == 0) {
		msg_box (dlg, _("No key was selected, select all by default."), _("Key Manager"), MB_INFO);
		listview_select_all (kmi->lv);
	    }
            km_refresh_from_keyserver (kmi->lv, dlg);
            break;
                        
	/** Context menu **/
	case ID_KEYCTX_EDIT:
	    send_cmd_id( dlg, ID_KEYMISC_EDITKEY );
	    break;

        case ID_KEYCTX_PROPS:
	    send_cmd_id( dlg, ID_KEYMISC_PROPS );
	    break;

        case ID_KEYCTX_UIDS: 
	    send_cmd_id( dlg, ID_KEYMISC_UID ); 
	    break;

        case ID_KEYCTX_SIGN: 
	    send_cmd_id( dlg, ID_KEYMISC_SIGN ); 
	    break;

        case ID_KEYCTX_DEL: 
	    send_cmd_id (dlg, ID_KEYMISC_DELETE); 
	    break;

        case ID_KEYCTX_REV: 
	    send_cmd_id( dlg, ID_KEYMISC_REVCERT ); 
	    break;

	case ID_KEYCTX_SENDMAIL:
	    km_send_to_mail_recipient (kmi->lv, dlg);
	    break;

        case ID_KEYCTX_KS_DEFAULT:
	    host = kserver_get_hostname (0, -1, &port);
	    km_send_to_keyserver (kmi->lv, dlg, host, port);
	    break;

	case ID_KEYCTX_ADDKEY:
            listview_get_item_text (kmi->lv, kmi->lv_idx, 1, keyid, DIM (keyid)-1);
	    memset (&k, 0, sizeof (k));
            k.key_pair = km_check_for_seckey (kmi->lv, kmi->lv_idx, NULL);
	    k.is_protected = km_check_if_protected (kmi->lv, kmi->lv_idx);
	    k.keyid = keyid;
	    keyedit_add_subkey (&k, dlg, NULL);
	    if (k.update)
		update_key (kmi->lv, kmi->lv_idx, keyid, 1);
	    break;

	case ID_KEYCTX_ADDUID:
	    listview_get_item_text( kmi->lv, kmi->lv_idx, 1, keyid, DIM (keyid)-1);
	    memset (&k, 0, sizeof (k));
	    k.key_pair = km_check_for_seckey (kmi->lv, kmi->lv_idx, NULL);
	    k.is_protected = km_check_if_protected (kmi->lv, kmi->lv_idx);
	    k.keyid = keyid;
	    keyedit_add_userid (&k, dlg, NULL);
	    if (k.update)
		update_key (kmi->lv, kmi->lv_idx, keyid, 1);
	    break;

	case ID_KEYCTX_ADDREV:
	    listview_get_item_text (kmi->lv, kmi->lv_idx, 1, keyid, DIM (keyid)-1);
	    memset (&k, 0, sizeof (k));
	    k.keyid = keyid;
	    k.is_protected = km_check_if_protected (kmi->lv, kmi->lv_idx);
	    k.key_pair = km_check_for_seckey( kmi->lv, kmi->lv_idx, NULL );
	    keyedit_add_revoker (&k, dlg);
	    if (k.update)
		update_key (kmi->lv, kmi->lv_idx, keyid, 1);
	    break;

	case ID_KEYCTX_ADDPHOTO:
	    listview_get_item_text (kmi->lv, kmi->lv_idx, 1, keyid, DIM (keyid)-1);
	    memset (&k, 0, sizeof (k));
	    k.keyid = keyid;
	    k.is_protected = km_check_if_protected (kmi->lv, kmi->lv_idx);
	    k.key_pair = km_check_for_seckey (kmi->lv, kmi->lv_idx, NULL);
	    keyedit_add_photo (&k, dlg);
	    if (k.update)
		update_key (kmi->lv, kmi->lv_idx, keyid, 1);
	    break;

	case ID_KEYCTX_KS_NL:
	case ID_KEYCTX_KS_PL:
	case ID_KEYCTX_KS_AT:
	case ID_KEYCTX_KS_DE:
	case ID_KEYCTX_KS_DK:
	case ID_KEYCTX_KS_CZ:
	case ID_KEYCTX_KS_ES:
	case ID_KEYCTX_KS_UK:
	    host = kserver_get_hostname (LOWORD (wparam) - 40107, 0, &port);
	    km_send_to_keyserver (kmi->lv, dlg, host, port);
	    break;

        case ID_KEYCTX_RECVFROM: 
	    km_refresh_from_keyserver (kmi->lv, dlg);
	    break;

	case ID_KEYCTX_UID_COPY:
	    /* XXX: add generic function to support multiple selection
	            with a callback */
	    listview_get_item_text( kmi->lv, kmi->lv_idx, 0, uid, sizeof uid-1 );
	    set_clip_text( NULL, uid, strlen( uid ) );
	    break;

	case ID_KEYCTX_KEYID_COPY:
	    listview_get_item_text (kmi->lv, kmi->lv_idx, 1, uid, sizeof uid-1);
	    set_clip_text (NULL, uid, strlen (uid));
	    break;

	case ID_KEYCTX_FPR_COPY:
	    key = (gpgme_key_t) listview_get_item2 (kmi->lv, kmi->lv_idx);	  
	    if (key) {
		t = get_key_fpr (key);
		set_clip_text (NULL, t? t : "", t? strlen (t): 0);
	    }
	    break;

	case ID_KEYCTX_KINFO_COPY:
	    listview_get_item_text( kmi->lv, kmi->lv_idx, 1, uid, sizeof uid-1 );
	    km_set_clip_info( uid );	    
	    break;

        case ID_KEYCTX_COPY:
            km_clip_export (dlg, kmi->lv);
            break;

        case ID_KEYCTX_PASTE:
            km_clip_import (dlg);
            break;

	case ID_KEYCTX_DISABLE:
	case ID_KEYCTX_ENABLE:
	    i = LOWORD (wparam) == ID_KEYCTX_ENABLE? 1 : 0;
	    listview_get_item_text (kmi->lv, kmi->lv_idx, KM_COL_KEYID, 
				    keyid, DIM (keyid)-1);
	    rc = km_enable_disable_key (kmi->lv, dlg, kmi->lv_idx, i);
	    if (!rc)
		update_key (kmi->lv, kmi->lv_idx, keyid, 0);
	    /* XXX: switching a key from disabled -> enabled. does not
	       change the validity field in the KM. */
	    break;

        case ID_KEYCTX_LISTSIGS: 
	    send_cmd_id (dlg, ID_KEYMISC_CHECKSIGS);
	    break;

	case ID_KEYCTX_MAXTRUST:
	    listview_get_item_text (kmi->lv, kmi->lv_idx, 1, keyid, DIM (keyid)-1);
	    rc = km_set_implicit_trust (dlg, kmi->lv, kmi->lv_idx);
	    if (!rc)
		update_key (kmi->lv, kmi->lv_idx, keyid, 0);
	    break;

        case ID_KEYCTX_SETDEFKEY:
	    if (!km_check_key_status (kmi->lv, kmi->lv_idx)) {
		listview_get_item_text (kmi->lv, kmi->lv_idx, 1, keyid, DIM (keyid)-1);
		rc = set_gnupg_default_key (keyid);
		if (rc)
		    msg_box (dlg, winpt_strerror (rc), _("Key Manager"), MB_ERR);
		km_update_default_key_str (kmi->statbar);
	    }
            break;

	case ID_GROUP_NEW:
        case ID_GROUP_PASTE:
        case ID_GROUP_DELETE:
	    /* XXX: Implement group manager. */
            return TRUE;
        }
	
        break;
    }
    
    return FALSE;
}



